"""Support code for different states of nodes"""

# -- Imports ------------------------------------------------------------------

from bidict import bidict

# -- Classes ------------------------------------------------------------------


class State:
    """Wrapper class for the byte returned by the Get/Set State command

    See also: https://mytoolit.github.io/Documentation/#command:get-set-state

    Args:

        value:
            The value of the state byte

        mode:
            Specifies if the state should be set or retrieved (get)

        location:
            A string or number that specifies the code location

        state:
            A string or number that specifies the network state

    Examples:

        Create some example state objects

        >>> State(mode='Get', location='Bootloader', state='Operating')
        Get State, Location: Bootloader, State: Operating

        >>> State(mode='Set', state='Operating').value == 0b1_0000_101
        True

    """

    node_states = bidict({
        "No Change": 0,
        "Bootloader": 1,
        "Application": 2,
        "Reserved": 3,
    })

    network_states = bidict({
        "Failure": 0,
        "Error": 1,
        "Standby": 2,
        "Graceful Degradation 2": 3,
        "Graceful Degradation 1": 4,
        "Operating": 5,
        "Startup": 6,
        "No Change": 7,
    })

    def __init__(
        self,
        *value: int,
        mode: str | None = None,
        location: None | int | str = None,
        state: None | int | str = None,
    ) -> None:

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            state_ones = 0xFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ state_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set bits to given value
            self.value |= number << start

        self.value = value[0] if value else 0
        cls = type(self)

        # ========
        # = Mode =
        # ========

        if mode is not None:
            if mode not in {"Get", "Set"}:
                raise ValueError(f"Unknown mode “{mode}”")

            set_part(start=7, width=1, number=int(mode == "Set"))

        # ============
        # = Location =
        # ============

        if isinstance(location, str):
            try:
                location = cls.node_states[location]
            except KeyError as error:
                raise ValueError(f"Unknown location “{location}”") from error

        if location is not None:
            set_part(start=4, width=2, number=location)

        # =========
        # = State =
        # =========

        if isinstance(state, str):
            try:
                state = cls.network_states[state]
            except KeyError as error:
                raise ValueError(f"Unknown state “{state}”") from error

        if state is not None:
            set_part(start=0, width=3, number=state)

    def __eq__(self, other: object) -> bool:
        """Compare this state to another object

        Args:

            other:
                The other object this state should be compared to

        Returns:

            - ``True``, if the given object is a state and it has the same
              value as this state

            - ``False``, otherwise

        Examples:

            Compare some example state objects

            >>> state1 = State(mode='Get',
            ...                location='Bootloader',
            ...                state='Operating')
            >>> state2 = State(mode='Set',
            ...                location='Bootloader',
            ...                state='Operating')

            >>> state1 == state2
            False

            >>> state1 == State(state2.value, mode='Get')
            True

        """

        if isinstance(other, State):
            return self.value == other.value

        return False

    def is_set(self) -> bool:
        """Check if the status should be set

        If this method returns ``False``, then the state should be retrieved
        (get) instead.

        Examples:

            Check if the network state should be set/get for example states

            >>> State(0b1_0000000).is_set()
            True
            >>> State(0b0_0000000).is_set()
            False
            >>> State(mode='Get').is_set()
            False
            >>> State(mode='Set').is_set()
            True

        """

        return bool((self.value >> 7) & 1)

    def location_name(self) -> str:
        """Retrieve the name of the current (code) location

        Returns:

            The name of the node state represented by this state object

        Examples:

            Get the location name for some example states

            >>> State(0b00_1010).location_name()
            'No Change'
            >>> State(0b01_1010).location_name()
            'Bootloader'
            >>> State(location='Application').location_name()
            'Application'

        """

        location = (self.value >> 4) & 0b11
        cls = type(self)
        # pylint: disable=unsubscriptable-object
        return cls.node_states.inverse[location]
        # pylint: enable=unsubscriptable-object

    def state_name(self) -> str:
        """Retrieve the name of the network state

        Returns:

            The name of the network state represented by this state object

        Examples:

            Get the state name for some example states

            >>> State(0b000).state_name()
            'Failure'
            >>> State(0b101).state_name()
            'Operating'
            >>> State(state='Startup').state_name()
            'Startup'

        """

        network_state = self.value & 0b111
        cls = type(self)
        # pylint: disable=unsubscriptable-object
        return cls.network_states.inverse[network_state]
        # pylint: enable=unsubscriptable-object

    def __repr__(self) -> str:
        """Retrieve the textual representation of the state

        Returns:

            A string that describes the attributes of the state

        Examples:

            Get the string representation of some example states

            >>> State(0b1_0_01_0_110)
            Set State, Location: Bootloader, State: Startup
            >>> State(0b0_1_11_1_001)
            Get State, Location: Reserved, State: Error

        """

        attributes = [
            f"{'Set' if self.is_set() else 'Get'} State",
            f"Location: {self.location_name()}",
            f"State: {self.state_name()}",
        ]

        return ", ".join(attributes)


class NodeStatus:
    """Wrapper class for the node status word

    Please do not use this class directly, but instead use one of the
    two specific status classes for the STH and STU.

    Args:

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            node status word


    """

    def __init__(self, value: list[int] | int) -> None:

        # Currently only the first byte (of the little endian version) of
        # status word 0 contains (non-reserved) data
        self.value = value if isinstance(value, int) else value[0]

    def __repr__(self) -> str:
        """Retrieve the textual representation of the node status word

        Returns:

            A string that describes the attributes of the node status word

        Examples:

            Get the textual representation of some node status words

            >>> NodeStatus(0b1010)
            State: Operating, No Error

            >>> NodeStatus([0b1010, 0, 0, 0])
            State: Operating, No Error

            >>> NodeStatus(0b1)
            State: Failure, Error

        """

        attributes = [
            f"State: {self.state_name()}",
            f"{'' if self.error() else 'No '}Error",
        ]

        return ", ".join(attributes)

    def error(self) -> bool:
        """Retrieve the status of the error bit

        Returns:

            ``True`` if the error bit was set or ``False`` otherwise

        Examples:

            Check if the error bit was set or not for some example status words

            >>> NodeStatus(0b0).error()
            False

            >>> NodeStatus(0b1).error()
            True

        """

        return bool(self.value & 1)

    def state_name(self) -> str:
        """Get the name of the state represented by the node status word

        Returns:

            A textual representation of the current node state

        Examples:

            Get the textual representation of some example node status words

            >>> NodeStatus(0b1010).state_name()
            'Operating'

            >>> NodeStatus(0b1110).state_name()
            'No Change'

        """

        state = (self.value >> 1) & 0b111
        # pylint: disable=unsubscriptable-object
        return State.network_states.inverse[state]


class NodeStatusSTH(NodeStatus):
    """Wrapper for the node status word of the STH"""


class NodeStatusSTU(NodeStatus):
    """Wrapper for the node status word of the STU"""

    def __repr__(self) -> str:
        """Retrieve the textual representation of the node status word

        Returns:

            A string that describes the attributes of the node status word

        Examples:

            Get the string representation of an STU node status word

            >>> NodeStatusSTU(0b1101010) # doctest:+NORMALIZE_WHITESPACE
            State: Operating, No Error, Radio Port Disabled,
            CAN Port Enabled, Bluetooth Connected

        """

        radio_port_enabled = self.value >> 4 & 1
        can_port_enabled = self.value >> 5 & 1
        bluetooth_connected = self.value >> 6 & 1

        attributes = [
            super().__repr__(),
            f"Radio Port {'Enabled' if radio_port_enabled else 'Disabled'}",
            f"CAN Port {'Enabled' if can_port_enabled else 'Disabled'}",
            (
                "Bluetooth "
                f"{'Connected' if bluetooth_connected else 'Disconnected'}"
            ),
        ]

        return ", ".join(attributes)


# pylint: disable=too-few-public-methods


class ErrorStatus:
    """Wrapper class for the error status word

    Please do not use this class directly, but instead use one of the
    two specific error status classes for the STH and STU.

    Args:

        value:
            A 32 bit integer or list of bytes that specifies the value of
            the error status word

    """

    def __init__(self, value: list[int] | int) -> None:

        # Currently only the first byte (of the little endian version) of
        # status word contains (non-reserved) data
        self.value = value if isinstance(value, int) else value[0]

    def transmission_error(self) -> bool:
        """Retrieve the status of the transmission error bit

        Returns:

            True if the error bit was set or False otherwise

        Examples:

            Check some error status words for transmission errors

            >>> ErrorStatus(0b0).transmission_error()
            False

            >>> ErrorStatus(0b1).transmission_error()
            True

        """

        return bool(self.value & 1)


# pylint: enable=too-few-public-methods


class ErrorStatusSTH(ErrorStatus):
    """Wrapper class for error status word 1 of the STH"""

    def __repr__(self) -> str:
        """Retrieve the textual representation of the error status word

        Returns:

            A string that describes the attributes of the error status word

        Examples:

            Get the textual representation of some STH error status words

            >>> ErrorStatusSTH(0b0)
            No Error

            >>> ErrorStatusSTH(0b11)
            Bluetooth Transmission Error, ADC Overrun Error

            >>> ErrorStatusSTH(0b10)
            ADC Overrun Error

        """

        errors = []

        if self.transmission_error():
            errors.append("Bluetooth Transmission Error")

        if self.adc_overrun():
            errors.append("ADC Overrun Error")

        if not errors:
            return "No Error"

        return ", ".join(errors)

    def adc_overrun(self) -> bool:
        """Retrieve the status of the ADC overrun bit

        Returns:

            ``True`` if the ADC overrun error bit is set or ``False`` otherwise

        Examples:

            Check some STH error status words for overrun errors

            >>> ErrorStatusSTH(0b10).adc_overrun()
            True

            >>> ErrorStatusSTH(0b11).adc_overrun()
            True

            >>> ErrorStatusSTH(0b01).adc_overrun()
            False

        """

        return bool((self.value >> 1) & 1)


class ErrorStatusSTU(ErrorStatus):
    """Wrapper class for error status word 1 of the STH"""

    def __repr__(self) -> str:
        """Retrieve the textual representation of the error status word

        Returns:

            A string that describes the attributes of the error status word

        Examples:

            Get the string representation of some STU status error words

            >>> ErrorStatusSTU(0b0)
            No Error

            >>> ErrorStatusSTU(0b1)
            CAN Transmission Error

        """

        if self.transmission_error():
            return "CAN Transmission Error"

        return "No Error"


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for storing measurement data (in HDF5)"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from datetime import datetime
from pathlib import Path
from types import TracebackType

from tables import (
    File,
    Filters,
    Float32Col,
    IsDescription,
    MetaAtom,
    MetaIsDescription,
    NoSuchNodeError,
    open_file,
    UInt8Col,
    UInt64Col,
)
from tables.exceptions import HDF5ExtError

from icotronic.can.adc import ADCConfiguration
from icotronic.can.streaming import StreamingConfiguration, StreamingData

# -- Functions ----------------------------------------------------------------


def create_acceleration_description(
    attributes: dict[str, MetaAtom],
) -> MetaIsDescription:
    """Create a new ``IsDescription`` class to store acceleration data

    Args:

        attributes:
            A dictionary containing additional columns to store specific
            acceleration data. The key specifies the name of the attribute,
            while the value specifies the type.

    Examples:

        Create an example acceleration description class

        >>> description_x_acceleration = create_acceleration_description(
        ...     dict(x=Float32Col()))
        >>> list(description_x_acceleration.columns.keys())
        ['counter', 'timestamp', 'x']

    """

    description_class = type(
        "ExtendedAccelerationDescription",
        (AccelerationDescription,),
        attributes,
    )
    return description_class


# -- Classes ------------------------------------------------------------------


# pylint: disable=too-few-public-methods


class AccelerationDescription(IsDescription):
    """Description of HDF acceleration table"""

    counter = UInt8Col()
    """Message counter"""

    timestamp = UInt64Col()
    """Microseconds since measurement start"""


# pylint: enable=too-few-public-methods


class StorageException(Exception):
    """Exception for HDF storage errors"""


class Storage:
    """Context manager class for storing measurement data in HDF5 format

    Args:

        filepath:
            The filepath of the HDF5 file in which this object should store
            measurement data

        channels:
            All channels for which data should be collected or ``None``, if the
            axes data should be taken from an existing valid file at
            ``filepath``.

    Examples:

        Create new file

        >>> filepath = Path("test.hdf5")
        >>> with Storage(filepath,
        ...              channels=StreamingConfiguration(first=True)
        ... ) as storage:
        ...     pass

        Opening an existing file but still providing channel info should fail

        >>> with Storage(filepath,
        ...             channels=StreamingConfiguration(first=True)
        ... ) as storage:
        ...     pass # doctest:+ELLIPSIS
        Traceback (most recent call last):
            ...
        ValueError: File “...” exist but channels parameter is not None
        >>> filepath.unlink()

    """

    def __init__(
        self,
        filepath: Path | str,
        channels: StreamingConfiguration | None = None,
    ) -> None:

        self.filepath = Path(filepath).expanduser().resolve()

        if channels and self.filepath.exists():
            raise ValueError(
                f"File “{self.filepath}” exist but channels parameter "
                "is not None"
            )

        self.hdf: File | None = None
        self.channels = channels

    def __enter__(self) -> StorageData:
        """Open the HDF file for writing"""

        return self.open()

    def __exit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Clean up the resources used by the storage class

        Args:

            exception_type:
                The type of the exception in case of an exception

            exception_value:
                The value of the exception in case of an exception

            traceback:
                The traceback in case of an exception

        """

        self.close()

    def open(self) -> StorageData:
        """Open and initialize the HDF file for writing"""

        try:
            self.hdf = open_file(
                self.filepath,
                mode="a",
                filters=Filters(4, "zlib"),
                title="STH Measurement Data",
            )
        except (HDF5ExtError, OSError) as error:
            raise StorageException(
                f"Unable to open file “{self.filepath}”: {error}"
            ) from error

        return StorageData(self.hdf, self.channels)

    def close(self) -> None:
        """Close the HDF file"""

        if isinstance(self.hdf, File) and self.hdf.isopen:
            self.hdf.close()


class StorageData:
    """Store HDF acceleration data

    Args:

        file_handle:
            The HDF file that should store the data

        channels:
            All channels for which data should be collected or ``None``, if the
            axes data should be taken from an existing valid file at
            ``filepath``.

    Examples:

        Create new data

        >>> filepath = Path("test.hdf5")
        >>> streaming_data = StreamingData(values=[1, 2, 3], counter=1,
        ...                                timestamp=4306978.449)
        >>> with Storage(filepath, StreamingConfiguration(first=True)) as data:
        ...     data.add_streaming_data(streaming_data)

        Read from existing file

        >>> with Storage(filepath) as data:
        ...     print(data.dataloss_stats())
        (1, 0)

        >>> filepath.unlink()

    """

    def __init__(
        self,
        file_handle: File,
        channels: StreamingConfiguration | None = None,
    ) -> None:

        self.hdf = file_handle
        self.start_time: float | None = None

        name = "acceleration"
        if channels:
            self.streaming_configuration = channels
            self.axes = channels.axes()
            self.acceleration = self.hdf.create_table(
                self.hdf.root,
                name=name,
                description=create_acceleration_description(
                    attributes={axis: Float32Col() for axis in self.axes}
                ),
                title="STH Acceleration Data",
            )
        else:
            try:
                self.acceleration = self.hdf.get_node(f"/{name}")
                streaming_config = {
                    "first": False,
                    "second": False,
                    "third": False,
                }
                for axis, channel in zip("xyz", streaming_config.keys()):
                    try:
                        getattr(self.acceleration.description, axis)
                        streaming_config[channel] = True
                    except AttributeError:
                        pass

                self.streaming_configuration = StreamingConfiguration(
                    **streaming_config
                )
                self.axes = self.streaming_configuration.axes()
                self.start_time = self.acceleration[-1][1] / 1000

            except NoSuchNodeError as error:
                raise StorageException(
                    f"Unable to open file “{self.hdf.filename}” due to "
                    f"incorrect format: {error}"
                ) from error

    def add_streaming_data(
        self,
        streaming_data: StreamingData,
    ) -> None:
        """Add streaming data to the storage object

        Args:

            streaming_data:
                The streaming data that should be added to the storage

        Examples:

            Import required library code

            >>> from icotronic.can.streaming import StreamingConfigBits

            Store streaming data for single channel

            >>> channel3 = StreamingConfiguration(first=False, second=False,
            ...                                   third=True)
            >>> data1 = StreamingData(values=[1, 2, 3], counter=21,
            ...                       timestamp=1)
            >>> data2 = StreamingData(values=[4, 5, 6], counter=22,
            ...                       timestamp=2)
            >>> filepath = Path("test.hdf5")
            >>> with Storage(filepath, channel3) as storage:
            ...     storage.add_streaming_data(data1)
            ...     storage.add_streaming_data(data2)
            ...     # Normally the class takes care about when to store back
            ...     # data to the disk itself. We do a manual flush here to
            ...     # check the number of stored items.
            ...     storage.acceleration.flush()
            ...     print(storage.acceleration.nrows)
            6
            >>> filepath.unlink()

            Store streaming data for three channels

            >>> all = StreamingConfiguration(first=True, second=True,
            ...                              third=True)
            >>> data1 = StreamingData(values=[1, 2, 3], counter=21,
            ...                       timestamp=1)
            >>> data2 = StreamingData(values=[4, 5, 6], counter=22,
            ...                       timestamp=2)
            >>> with Storage(filepath, all) as storage:
            ...     storage.add_streaming_data(data1)
            ...     storage.add_streaming_data(data2)
            ...     storage.acceleration.flush()
            ...     print(storage.acceleration.nrows)
            2
            >>> filepath.unlink()

        """

        values = streaming_data.values
        timestamp = streaming_data.timestamp
        counter = streaming_data.counter

        if self.start_time is None:
            self.start_time = timestamp
            self.acceleration.attrs["Start_Time"] = datetime.now().isoformat()

        assert isinstance(self.start_time, (int, float))

        row = self.acceleration.row
        timestamp = (timestamp - self.start_time) * 1_000_000

        if len(self.axes) == 1:
            axis = self.axes[0]
            for value in values:
                row["timestamp"] = timestamp
                row["counter"] = counter
                row[axis] = value
                row.append()
        else:
            row["timestamp"] = timestamp
            row["counter"] = counter
            for accelertation_type, value in zip(self.axes, values):
                row[accelertation_type] = value
            row.append()

    def store_acceleration_meta(self, name: str, value: str) -> None:
        """Add acceleration metadata

        Args:

            name:
                The name of the meta attribute

            value:
                The value of the meta attribute

        Examples:

            Store some example acceleration data

            >>> filepath = Path("test.hdf5")
            >>> with Storage(filepath,
            ...              StreamingConfiguration(third=True)) as storage:
            ...     storage.store_acceleration_meta("something", "some value")
            ...     print(storage.acceleration.attrs["something"])
            some value
            >>> filepath.unlink()

        """

        self.acceleration.attrs[name] = value

    def write_sensor_range(self, sensor_range_in_g: float) -> None:
        """Add metadata about sensor range

        This method assumes that sensor have a symmetric measurement range
        (e.g. a sensor with a range of 200 g measures from - 100 g up to
        + 100 g).

        Args:

            sensor_range_in_g:
                The measurement range of the sensor in multiples of g

        Examples:

            Add sensor range metadata to example file

            >>> filepath = Path("test.hdf5")
            >>> with Storage(filepath,
            ...              StreamingConfiguration(third=True)) as storage:
            ...     storage.write_sensor_range(200)
            ...     print(storage.acceleration.attrs["Sensor_Range"])
            ± 100 g₀
            >>> filepath.unlink()

        """

        sensor_range_positive = round(sensor_range_in_g / 2)

        self.store_acceleration_meta(
            "Sensor_Range", f"± {sensor_range_positive} g₀"
        )

    def write_sample_rate(self, adc_configuration: ADCConfiguration) -> None:
        """Store the sample rate of the ADC

        Args:

            adc_configuration:
                The current ADC configuration of the sensor node

        Examples:

            Write and read sample rate metadata in example HDF file

            >>> filepath = Path("test.hdf5")
            >>> adc_configuration = ADCConfiguration(
            ...     set=True,
            ...     prescaler=2,
            ...     acquisition_time=16,
            ...     oversampling_rate=256)
            >>> with Storage(filepath,
            ...              StreamingConfiguration(first=True)) as storage:
            ...     storage.write_sample_rate(adc_configuration)
            ...     sample_rate = storage.acceleration.attrs["Sample_Rate"]
            ...     print(sample_rate) # doctest:+NORMALIZE_WHITESPACE
            1724.14 Hz (Prescaler: 2, Acquisition Time: 16,
            Oversampling Rate: 256)
            >>> filepath.unlink()

        """

        sample_rate = adc_configuration.sample_rate()

        adc_config_text = ", ".join([
            f"Prescaler: {adc_configuration.prescaler}",
            f"Acquisition Time: {adc_configuration.acquisition_time}",
            f"Oversampling Rate: {adc_configuration.oversampling_rate}",
        ])

        self.store_acceleration_meta(
            "Sample_Rate", f"{sample_rate:.2f} Hz ({adc_config_text})"
        )

    def dataloss_stats(self) -> tuple[int, int]:
        """Determine number of lost and received messages

        Returns:

            Tuple containing the number of received and the number of lost
            messages

        Examples:

            Get data loss statistics for an example file

            >>> def calculate_dataloss_stats():
            ...     filepath = Path("test.hdf5")
            ...     with Storage(filepath,StreamingConfiguration(
            ...                  first=True)) as storage:
            ...         for counter in range(256):
            ...             storage.add_streaming_data(
            ...                 StreamingData(values=[1, 2, 3],
            ...                               counter=counter,
            ...                               timestamp=counter/10))
            ...         for counter in range(128, 256):
            ...             storage.add_streaming_data(
            ...                 StreamingData(values=[4, 5, 6],
            ...                               counter=counter,
            ...                               timestamp=(255 + counter)/10))
            ...
            ...         stats = storage.dataloss_stats()
            ...     filepath.unlink()
            ...     return stats
            >>> retrieved, lost = calculate_dataloss_stats()
            >>> retrieved
            384
            >>> lost
            128

        """

        # Write back acceleration data so we can iterate over it
        self.acceleration.flush()

        lost_messages = 0
        last_counter = int(self.acceleration[0][0])

        for record in self.acceleration:
            counter = int(record[0])

            if counter == last_counter:
                continue  # Skip data with same message counter

            lost_messages += (counter - last_counter) % 256 - 1

            last_counter = counter

        number_rows = len(self.acceleration)
        # 3 axes → 1 message ↔ 1 row
        # 2 axes → 1 message ↔ 1 row
        # 1 axis → 1 message ↔ 3 rows

        retrieved_messages = (
            number_rows if len(self.axes) >= 2 else number_rows // 3
        )

        return (retrieved_messages, lost_messages)

    def dataloss(self) -> float:
        """Determine (minimum) data loss

        Returns:

            Amount of lost messages divided by all messages (lost and
            retrieved)

        Examples:

            Import required library code

            >>> from math import isclose

            Calculate data loss for an example storage file

            >>> def calculate_dataloss():
            ...     filepath = Path("test.hdf5")
            ...     with Storage(filepath, StreamingConfiguration(
            ...                  first=True)) as storage:
            ...         for counter in range(256):
            ...             storage.add_streaming_data(
            ...                 StreamingData(values=[1, 2, 3],
            ...                               counter=counter,
            ...                               timestamp=counter/10))
            ...         for counter in range(128, 256):
            ...             storage.add_streaming_data(
            ...                 StreamingData(values=[4, 5, 6],
            ...                               counter=counter,
            ...                               timestamp=(255 + counter)/10))
            ...
            ...         dataloss = storage.dataloss()
            ...     filepath.unlink()
            ...     return dataloss
            >>> isclose(0.25, calculate_dataloss(), rel_tol=0.005)
            True

        """

        retrieved_messages, lost_messages = self.dataloss_stats()

        messages = retrieved_messages + lost_messages

        return lost_messages / messages if messages > 0 else 0

    def sampling_frequency(self) -> float:
        """Calculate sampling frequency of measurement data

        Returns:

            Sampling frequency (of a single data channel) in Hz

        Examples:

            Import required library code

            >>> from math import isclose

            Calculate sampling frequency for an example file

            >>> def calculate_sampling_frequency():
            ...     filepath = Path("test.hdf5")
            ...     with Storage(filepath, StreamingConfiguration(
            ...             first=True, second=True, third=True)) as storage:
            ...         storage.add_streaming_data(
            ...             StreamingData(values=[1, 2, 3],
            ...                           counter=1,
            ...                           timestamp=0))
            ...         storage.add_streaming_data(
            ...             StreamingData(values=[1, 2, 3],
            ...                           counter=2,
            ...                           timestamp=1))
            ...
            ...         sampling_frequency = storage.sampling_frequency()
            ...     filepath.unlink()
            ...     return sampling_frequency
            >>> calculate_sampling_frequency()
            2.0

        """

        retrieved_messages, lost_messages = self.dataloss_stats()
        messages = retrieved_messages + lost_messages
        rows_per_message = 3 if len(self.axes) == 1 else 1
        if len(self.acceleration) > 1:
            measurement_time_in_us = int(self.acceleration[-1]["timestamp"])
            return rows_per_message * messages * 10**6 / measurement_time_in_us

        return 0


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support code for acceleration measurements"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterable
from math import log, sqrt
from statistics import pvariance

from icotronic.measurement.constants import ADC_MAX_VALUE

# -- Functions ----------------------------------------------------------------


def convert_raw_to_g(acceleration_raw: float, max_value: float) -> float:
    """Convert an acceleration value sent by the STH into a factor

    The factor measures the amount of the gravitational force
    (g₀ = 9.807 m/s²) applied to the STH.

    Args:

        acceleration_raw:
            The 16 bit integer acceleration value as sent by the STH

        max_value:
            The maximum acceleration value as factor of g₀
            (e.g. 200 for a ±100 g sensor)

    Returns:

        The acceleration in multiples of the standard gravity g₀

    Examples:

        Convert a raw value (that represents ca. 0 g₀) into multiples of g₀

        >>> acceleration = convert_raw_to_g(2**15, max_value=200)
        >>> -0.01 < acceleration < 0.01
        True

    """

    acceleration_to_gravity = max_value

    # The code subtracts 1/2 from the computed value, since the STH linearly
    # maps the maximum negative acceleration to 0 and the maximum positive
    # acceleration to the maximum ADC value.
    acceleration_in_g = (
        acceleration_raw / ADC_MAX_VALUE - 1 / 2
    ) * acceleration_to_gravity
    return acceleration_in_g


def ratio_noise_max(values: Iterable[int]) -> float:
    """Calculate the ratio noise to max ADC amplitude in dB

    Args:

        values:
            An iterable object that stores a series of measured 16 bit raw ADC
            (acceleration) values

    Returns:

        The ratio of the average noise to the highest possible measured value

    """

    max_value = ADC_MAX_VALUE / 2
    standard_deviation = sqrt(pvariance(values))
    return 20 * log(standard_deviation / max_value, 10)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for changing configuration values

Currently the configuration is mainly used in the hardware (production tests).
"""

# -- Import -------------------------------------------------------------------

from datetime import date, datetime
from functools import partial
from importlib.resources import as_file, files
from numbers import Real
from os import makedirs
from pathlib import Path
from platform import system
from sys import exit as sys_exit, stderr

from dynaconf import Dynaconf, ValidationError, Validator
from dynaconf.vendor.ruamel.yaml.parser import ParserError
from dynaconf.vendor.ruamel.yaml.scanner import ScannerError
from platformdirs import site_config_dir, user_config_dir

from icotronic.utility.open import open_file, UnableToOpenError

# -- Classes ------------------------------------------------------------------


class ConfigurationUtility:
    """Access configuration data"""

    app_name = "ICOtronic"
    app_author = "MyTooliT"
    config_filename = "config.yaml"
    site_config_filepath = (
        Path(site_config_dir(app_name, appauthor=app_author)) / config_filename
    )
    user_config_filepath = (
        Path(user_config_dir(app_name, appauthor=app_author)) / config_filename
    )

    @staticmethod
    def open_config_file(filepath: Path):
        """Open configuration file

        Args:

            filepath:
                Path to configuration file

        """

        # Create file, if it does not exist already
        if not filepath.exists():
            filepath.parent.mkdir(
                exist_ok=True,
                parents=True,
            )

            default_user_config = (
                files("icotronic.config")
                .joinpath("user.yaml")
                .read_text(encoding="utf-8")
            )

            with filepath.open("w", encoding="utf8") as config_file:
                config_file.write(default_user_config)

        open_file(filepath)

    @classmethod
    def open_user_config(cls):
        """Open the current users configuration file"""

        try:
            cls.open_config_file(cls.user_config_filepath)
        except UnableToOpenError as error:
            print(
                f"Unable to open user configuration: {error}"
                "\nTo work around this problem please open "
                f"“{cls.user_config_filepath}” in your favorite text "
                "editor",
                file=stderr,
            )


class SettingsIncorrectError(Exception):
    """Raised when the configuration is incorrect"""


class Settings(Dynaconf):
    """Small extension of the settings object for our purposes

    Args:

        default_settings_filepath:
            Filepath to default settings file

        setting_files:
            A list containing setting files in ascending order according to
            importance (most important last).

        arguments:
            All positional arguments

        keyword_arguments:
            All keyword arguments

    """

    def __init__(
        self,
        default_settings_filepath,
        *arguments,
        settings_files: list[str] | None = None,
        **keyword_arguments,
    ) -> None:

        if settings_files is None:
            settings_files = []

        # If we use PyInstaller to create an installer for the package, then
        # loading the default config from the package data (`config.yaml`) does
        # not seem to work. As a workaround we also load data from
        # `default.yaml` inside the user configuration directory. This way you
        # can copy the content of `config.yaml` to this location before
        # executing any code and they ICOtronic library should still work,
        # even when loading the default config from the package data fails.
        user_default_filepath = (
            Path(ConfigurationUtility.user_config_filepath).parent
            / "default.yaml"
        )
        settings_files = [
            default_settings_filepath,
            ConfigurationUtility.site_config_filepath,
            user_default_filepath,
            ConfigurationUtility.user_config_filepath,
        ] + settings_files

        super().__init__(
            settings_files=settings_files,
            *arguments,
            **keyword_arguments,
        )
        self.validate_settings()

    # pylint: disable=too-many-locals

    def validate_settings(self) -> None:
        """Check settings for errors"""

        def must_exist(*arguments, **keyword_arguments):
            """Return Validator which requires setting to exist"""

            return Validator(*arguments, must_exist=True, **keyword_arguments)

        def element_is_type(
            nodes,
            name: str,
            element_type: type,
        ):
            """Check that all elements of a list have a certain type"""
            if nodes is None:
                return True  # Let parent validator handle wrong type

            for node in nodes:
                if not isinstance(node, element_type):
                    raise ValidationError(
                        f"Element “{node}” of {name} has wrong type "
                        f"“{type(node)}” instead of “{element_type}”"
                    )
            return True

        def element_is_string(nodes, name: str):
            """Check that all elements of a list are strings"""
            return element_is_type(nodes, name, element_type=str)

        def element_is_int(nodes, name: str):
            """Check that all elements of a list are ints"""
            return element_is_type(nodes, name, element_type=int)

        def node_validators(name: str):
            """Return shared validator for ICOtronic node (STH, STU, SMH)"""

            return [
                must_exist(
                    f"{name}.batch_number",
                    f"{name}.gtin",
                    f"{name}.programming_board.serial_number",
                    is_type_of=int,
                ),
                must_exist(
                    f"{name}.firmware.location.flash",
                    f"{name}.firmware.release_name",
                    f"{name}.hardware_version",
                    is_type_of=str,
                ),
                must_exist(
                    f"{name}.oem_data",
                    is_type_of=list,
                    condition=partial(element_is_int, name="{name}.oem_data"),
                ),
                must_exist(
                    f"{name}.production_date",
                    is_type_of=date,
                ),
                must_exist(
                    f"{name}.product_name",
                    is_type_of=str,
                    len_max=128,
                    cast=str,
                ),
                must_exist(
                    f"{name}.serial_number",
                    is_type_of=str,
                    len_max=8,
                    cast=str,
                ),
            ]

        def sensor_node_validators(name: str):
            """Return shared validator for STH or SMH"""

            return node_validators(name) + [
                must_exist(
                    f"{name}.name",
                    is_type_of=str,
                ),
            ]

        def sensor_validators(name: str):
            prefix = "sth.acceleration_sensor"
            return [
                must_exist(
                    f"{prefix}.{name}.acceleration.maximum",
                    f"{prefix}.{name}.acceleration.ratio_noise_to_max_value",
                    f"{prefix}.{name}.acceleration.tolerance",
                    f"{prefix}.{name}.reference_voltage",
                    f"{prefix}.{name}.self_test.voltage.difference",
                    f"{prefix}.{name}.self_test.voltage.tolerance",
                    is_type_of=Real,
                ),
                must_exist(
                    f"{prefix}.{name}.self_test.dimension",
                    is_type_of=str,
                    is_in=("x", "y", "z"),
                ),
            ]

        config_system = "mac" if system() == "Darwin" else system().lower()
        can_validators = [
            must_exist(f"can.{config_system}.bitrate", is_type_of=int),
            must_exist(
                f"can.{config_system}.channel",
                f"can.{config_system}.interface",
                is_type_of=str,
            ),
        ]
        commands_validators = [
            must_exist(
                "commands.path.linux",
                is_type_of=list,
                condition=partial(
                    element_is_string, name="commands.path.linux"
                ),
            ),
            must_exist(
                "commands.path.mac",
                is_type_of=list,
                condition=partial(element_is_string, name="commands.path.mac"),
            ),
            must_exist(
                "commands.path.windows",
                is_type_of=list,
                condition=partial(
                    element_is_string, name="commands.path.windows"
                ),
            ),
        ]
        logger_validators = [
            must_exist(
                "logger.can.level",
                is_type_of=str,
                is_in=(
                    "CRITICAL",
                    "ERROR",
                    "WARNING",
                    "INFO",
                    "DEBUG",
                    "NOTSET",
                ),
            )
        ]
        measurement_validators = [
            must_exist(
                "measurement.output.directory",
                "measurement.output.filename",
                is_type_of=str,
            ),
        ]
        operator_validators = [must_exist("operator.name", is_type_of=str)]
        general_sensor_node_validators = [
            must_exist(
                "sensor_node.bluetooth.advertisement_time_1",
                "sensor_node.bluetooth.advertisement_time_2",
                "sensor_node.bluetooth.sleep_time_1",
                "sensor_node.bluetooth.sleep_time_2",
                is_type_of=int,
            )
        ]
        smh_validators = sensor_node_validators("smh") + [
            must_exist(
                "smh.channels",
                is_type_of=int,
            )
        ]
        sth_validators = (
            sensor_node_validators("sth")
            + sensor_validators("ADXL1001")
            + sensor_validators("ADXL1002")
            + sensor_validators("ADXL356")
        ) + [
            must_exist(
                "sth.acceleration_sensor.sensor",
                is_in=(
                    "ADXL1001",
                    "ADXL1002",
                    "ADXL356",
                ),
            ),
            must_exist(
                "sth.battery_voltage.average",
                "sth.battery_voltage.tolerance",
                is_type_of=Real,
            ),
            must_exist(
                "sth.status",
                is_in=(
                    "Bare PCB",
                    "Epoxied",
                ),
            ),
        ]
        stu_validators = node_validators("stu")

        self.validators.register(
            *can_validators,
            *commands_validators,
            *logger_validators,
            *measurement_validators,
            *operator_validators,
            *general_sensor_node_validators,
            *smh_validators,
            *sth_validators,
            *stu_validators,
        )

        try:
            self.validators.validate()
        except ValidationError as error:
            config_files_text = "\n".join((
                f"  • {ConfigurationUtility.site_config_filepath}",
                f"  • {ConfigurationUtility.user_config_filepath}",
            ))
            raise SettingsIncorrectError(
                f"Incorrect configuration: {error}\n\n"
                "Please make sure that the configuration files:\n\n"
                f"{config_files_text}\n\n"
                "contain the correct configuration values"
            ) from error

    # pylint: enable=too-many-locals

    def acceleration_sensor(self):
        """Get the settings for the current acceleration sensor

        Returns:

            A configuration object for the currently selected accelerometer
            sensor

        """

        sensor_settings = self.sth.acceleration_sensor
        return sensor_settings[sensor_settings.sensor]

    def sth_name(self) -> str:
        """Return the current name of the STH as string

        Returns:

            The name of the STH

        """

        return str(self.sth.name)

    def output_directory(self) -> Path:
        """Get the HDF output directory

        Returns:

            The HDF output directory as path object

        """

        directory = Path(settings.measurement.output.directory)
        return directory if directory.is_absolute() else directory.expanduser()

    def get_output_filepath(self) -> Path:
        """Get filepath of HDF measurement file

        The filepath returned by this method will always include a current
        timestamp to make sure that there are no conflicts with old output
        files.

        Returns:

            The path to the current HDF file

        """

        directory = self.output_directory()
        filename = Path(settings.measurement.output.filename)

        if not filename.suffix:
            filename = filename.with_suffix(".hdf5")

        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        filepath = directory.joinpath(
            f"{filename.stem}_{timestamp}{filename.suffix}"
        )

        return filepath

    def check_output_directory(self) -> None:
        """Check the output directory

        If the directory does not already exist, then this function will
        try to create it.

        """

        directory = self.output_directory()

        if directory.exists() and not directory.is_dir():
            raise NotADirectoryError(
                f"The output directory “{directory}” points to an "
                "existing file not a directory"
            )

        if not directory.is_dir():
            try:
                makedirs(str(directory))
            except OSError as error:
                raise OSError(
                    "Unable to create the output directory "
                    f"“{directory}”: {error}"
                ) from error


# -- Attributes ---------------------------------------------------------------


with as_file(
    files("icotronic.config").joinpath(ConfigurationUtility.config_filename)
) as repo_settings_filepath:
    try:
        settings = Settings(default_settings_filepath=repo_settings_filepath)
    except SettingsIncorrectError as settings_incorrect_error:
        print(f"{settings_incorrect_error}", file=stderr)
        sys_exit(1)
    except (ParserError, ScannerError) as parsing_error:
        print(f"Unable to parse configuration: {parsing_error}", file=stderr)
        sys_exit(1)



"""Support for MyTooliT CAN commands

A command is a 16 bit subpart of a CAN identifier used by the MyTooliT
protocol. For more information, please take a look here:

https://mytoolit.github.io/Documentation/#command
"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from icotronic.can.protocol.blocks import (
    blocks,
    UnknownBlockError,
    UnknownBlockCommandError,
)

# -- Classes ------------------------------------------------------------------


class Command:
    """This class represents a command including error and acknowledge bits

    Usually you will either specify the command directly, or provide
    block, block command and values for the error and acknowledge/request
    bits. If you decide to specify both the command value and one of the
    keyword arguments, then the keyword arguments will be used to
    overwrite specific parts of the command. For more information, please
    take a look at the examples.

    Args:

        command:
            A 16 bit number that specifies the command including acknowledge
            and error bits (16 bit)

        block:
            A 6 bit number or string that specifies the block of the command

        block_command:
            A 8 bit number or string that specifies the block command

        request:
            A boolean value that specifies if the command is for a request or
            an acknowledgement

        error:
            A boolean value that defines if there was an error or not

    Examples:

        Create some valid commands

        >>> Command(block=0, block_command=0).value
        0

        >>> #           block  command  A E
        >>> command = 0b001000_00000100_0_0
        >>> bin(Command(command).value)
        '0b10000000010000'

        >>> command = Command(block='System')
        >>> command.block_name()
        'System'

        >>> command = Command(block='Streaming', block_command='Data')
        >>> command.block_command_name()
        'Data'

        Specifying an incorrect block will cause an error

        >>> Command(block='Does Not Exist')
        Traceback (most recent call last):
            ...
        ValueError: Unknown block: Does Not Exist

        Specifying an incorrect block command will cause an error

        >>> Command(block='Streaming', block_command='Does Not Exist')
        Traceback (most recent call last):
            ...
        ValueError: Unknown block command: Does Not Exist

    """

    def __init__(
        self,
        *command: int,
        block: None | str | int = None,
        block_command: None | str | int = None,
        request: bool | None = None,
        error: bool | None = None,
    ) -> None:

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            command_ones = 0xFFFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ command_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set command bits to given value
            self.value |= number << start

        # ===========
        # = Command =
        # ===========

        self.value = command[0] if command else 0

        # =========
        # = Block =
        # =========

        if isinstance(block, str):
            try:
                block = blocks[block].number
            except UnknownBlockError as exception:
                raise ValueError(f"Unknown block: {block}") from exception

        if block is not None:
            set_part(start=10, width=6, number=block)

        # =================
        # = Block Command =
        # =================

        if isinstance(block_command, str):
            try:
                block_command_names = blocks[self.block()]
            except UnknownBlockError as exception:
                raise ValueError(
                    f"Unknown block number: {block}"
                ) from exception

            try:
                block_command = block_command_names[block_command].number
            except UnknownBlockCommandError as exception:
                raise ValueError(
                    f"Unknown block command: {block_command}"
                ) from exception

        if block_command is not None:
            set_part(start=2, width=8, number=block_command)

        # ===================
        # = Request & Error =
        # ===================

        if request is not None:
            set_part(start=1, width=1, number=int(bool(request)))

        if error is not None:
            set_part(start=0, width=1, number=int(bool(error)))

    def __repr__(self) -> str:
        """Get a textual representation of the command

        Returns:

            A string that describes the various attributes of the command

        Examples:

            Get the string representation of some example commands

            >>> #         block  command  A E
            >>> Command(0b001000_00000100_0_0)
            Block: StatisticalData, Command: ProductionDate, Acknowledge

            >>> Command(block=0, block_command=0x0c, request=True, error=False)
            Block: System, Command: Routing, Request

        """

        error = self.value & 1

        attributes = filter(
            None,
            [
                f"Block: {self.block_name()}",
                f"Command: {self.block_command_name()}",
                "Acknowledge" if self.is_acknowledgment() else "Request",
                "Error" if error else None,
            ],
        )

        return ", ".join(attributes)

    def block(self) -> int:
        """Get the block

        Returns:

            The block number of the command

        Examples:

            Get the block number of a example command

            >>> #         block  command  A E
            >>> Command(0b000011_00000000_0_0).block()
            3

        """

        return (self.value >> 10) & 0b111111

    def block_name(self) -> str:
        """Get a short description of the block

        Returns:

            A short textual representation of the block number

        Examples:

            Get the block name of some example commands

            >>> #         block  command  A E
            >>> Command(0b101000_00000010_0_0).block_name()
            'Configuration'

            >>> #         block  command  A E
            >>> Command(0b111101_00000010_0_0).block_name()
            'EEPROM'

            Get the block name of an unknown block

            >>> #         block  command  A E
            >>> Command(0b100000_00000010_0_0).block_name()
            'Unknown'

        """

        try:
            return blocks[self.block()].name
        except UnknownBlockError:
            return "Unknown"

    def block_command(self) -> int:
        """Get the block command number

        Returns:

            The block command number of the command

        Examples:

            Get the block command number of an example command

            >>> #         block  command  A E
            >>> Command(0b001000_00000100_0_0).block_command()
            4

        """

        return (self.value >> 2) & 0xFF

    def block_command_name(self) -> str:
        """Get the name of the block command

        Returns:

            A short textual representation of the block command

        Examples:

            Get the block command name of some example commands

            >>> #         block  command  A E
            >>> Command(0b101000_00000000_0_0).block_command_name()
            'Get/Set ADC Configuration'

            >>> #         block  command  A E
            >>> Command(0b000000_00001011_0_0).block_command_name()
            'Bluetooth'

        """

        try:
            return blocks[self.block()][self.block_command()].name
        except UnknownBlockError:
            return "Unknown"

    def is_acknowledgment(self) -> bool:
        """Checks if this command represents an acknowledgment

        Returns:

            - ``True`` if the command is for an acknowledgement
            - ``False`` otherwise

        Examples:

            Check if some example commands represent an acknowledgment

            >>> #         block  command  A E
            >>> Command(0b101000_00000000_0_0).is_acknowledgment()
            True

            >>> Command(request=True).is_acknowledgment()
            False

        """

        return bool((self.value >> 1) & 1 == 0)

    def is_error(self) -> bool:
        """Checks if the command represents an error

        Returns:

            - ``True`` if the command represents an error
            - ``False`` otherwise

        Examples:

            Check if some example commands represent an error

            >>> #         block  command  A E
            >>> Command(0b101011_00000001_1_0).is_error()
            False

            >>> #         block  command  A E
            >>> Command(0b101010_00000000_0_1).is_error()
            True

        """

        return bool(self.value & 1)

    def set_acknowledgment(self, value: bool = True) -> Command:
        """Set the acknowledgment bit to the given value

        Parameters:

            value:
                A boolean that specifies if the command represents an
                acknowledgment or not

        Returns:

            The modified command object

        Examples:

            Check if some example commands represent an acknowledgment

            >>> Command().set_acknowledgment().is_acknowledgment()
            True

            >>> Command().set_acknowledgment(True).is_acknowledgment()
            True

            >>> Command().set_acknowledgment(False).is_acknowledgment()
            False

        """

        request = not value
        request_bit = 1 << 1

        if request:
            self.value |= request_bit
        else:
            command_ones = 0xFF
            self.value &= request_bit ^ command_ones

        return self

    def set_error(self, error: bool = True) -> Command:
        """Set the error bit to the given value

        Args:

            error:
                A boolean that specifies if the command represents an error or
                not

        Returns:

            The modified command object

        Examples:

            Set the error bit of some commands and check the result

            >>> Command().set_error().is_error()
            True

            >>> Command().set_error(True).is_error()
            True

            >>> Command().set_error(False).is_error()
            False

        """

        if error:
            self.value |= 1
        else:
            self.value &= 0b11111110

        return self


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Test code for stationary transceiver unit (STU)"""

# -- Imports ------------------------------------------------------------------

import sys
import threading

from asyncio import sleep
from os import devnull
from sys import stderr
from unittest import main as unittest_main

from icotronic.can.connection import Connection
from icotronic.config import settings
from icotronic.report.report import Report
from icotronic.test.unit import ExtendedTestRunner
from icotronic.can.error import CANInitError, NoResponseError
from icotronic.test.production.node import BaseTestCases

# -- Class --------------------------------------------------------------------


class TestSTU(BaseTestCases.TestNode):
    """This class contains tests for the Stationary Transceiver Unit (STU)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()
        cls.report = Report(node="STU")

    async def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into STU

        Note:

            The additional underscore in the method name that makes sure this
            test case is executed before all other test cases.

        Note:

            The text ``disconnected`` in the method name makes sure that the
            test framework does not initialize a connection.

        """

        self._test_firmware_flash(
            flash_location=settings.stu.firmware.location.flash,
            programmmer_serial_number=(
                settings.stu.programming_board.serial_number
            ),
            chip="BGM111A256V2",
        )

        # Try to Fix CAN connection problems after firmware update

        # Ignore errors in notifier thread
        old_excepthook = threading.excepthook
        threading.excepthook = lambda *args, **kwargs: print("", end="")

        old_stderr = sys.stderr
        with open(devnull, "w", encoding="utf8") as stream_devnull:
            # Do not print error output
            sys.stderr = stream_devnull

            status_ok = False
            attempt = 1
            while not status_ok and attempt <= 10:
                print(
                    f"\nTrying to fix CAN connection (Attempt {attempt})",
                    end="",
                )
                try:
                    async with Connection() as stu:
                        await stu.reset()
                    status_ok = True

                except CANInitError:
                    # Init error only seems to happen on the **first
                    # attempt**, if the CAN adapter is not connected to the
                    # computer.
                    if attempt == 1:
                        print("\nCAN adapter is not connected → Exiting\n")
                        return

                    await sleep(1)
                except NoResponseError:
                    await sleep(1)

                attempt += 1
            print()

        # Reenable error output
        sys.stderr = old_stderr
        # Show errors in notifier threads again
        threading.excepthook = old_excepthook

        if not status_ok:
            print("Unable to fix CAN connection", file=stderr)

    async def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        # ================
        # = Product Data =
        # ================

        await self._test_eeprom_product_data(settings.stu)

        # ==============
        # = Statistics =
        # ==============

        await self._test_eeprom_statistics(
            settings.stu.production_date,
            settings.stu.batch_number,
        )

        # =================
        # = EEPROM Status =
        # =================

        await self._test_eeprom_status()


# -- Main ---------------------------------------------------------------------


def main():
    """Run production test for Stationary Transceiver Unit (STU)"""

    unittest_main(
        testRunner=ExtendedTestRunner, module="icotronic.test.production.stu"
    )


if __name__ == "__main__":
    main()



"""Support for a node (single CAN communication unit) in the ICOtronic system

See: https://mytoolit.github.io/Documentation/#mytoolit-communication-protocol
"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from re import fullmatch

# -- Classes ------------------------------------------------------------------


class NodeId:
    """This class represents a CAN node of the ICOtronic system

    A node represents a communication participant, such as a specific STH
    or a specific STU in the ICOtronic system.

    Args:

        node:
            The number that identifies this node

    Examples:

        Create node with default value

        >>> NodeId().value
        0

        Create nodes with string values

        >>> NodeId('STH 1')
        STH 1

        >>> NodeId('STU1')
        STU 1

        >>> NodeId('SPU 1').value
        15

        Using incorrect numbers to initialize a node will fail

        >>> NodeId('STU 15')
        Traceback (most recent call last):
           ...
        ValueError: Unknown node identifier “STU 15”

        You can create a copy of an already existing node

        >>> NodeId(NodeId('STH 2'))
        STH 2

    """

    def __init__(self, node: int | str | NodeId = 0) -> None:

        if isinstance(node, str):
            # Check for broadcast pseudo nodes
            broadcast_match = fullmatch(
                "Broadcast With(?P<no_acknowledgment>out)? Acknowledgment",
                node,
            )
            if broadcast_match:
                self.value = 0 if broadcast_match["no_acknowledgment"] else 31
                return

            # Check normal nodes
            node_match = fullmatch(
                r"(?P<name>S(?:PU|TH|TU)) ?(?P<number>\d{1,2})", node
            )

            if node_match is not None:
                node_name = node_match["name"]
                node_number = int(node_match["number"])

                if node_name == "STH" and 1 <= node_number <= 14:
                    self.value = node_number
                    return

                if node_name == "SPU" and 1 <= node_number <= 2:
                    self.value = node_number + 14
                    return

                if node_name == "STU" and 1 <= node_number <= 14:
                    self.value = node_number + 16
                    return

            raise ValueError(f"Unknown node identifier “{node}”")

        if isinstance(node, NodeId):
            self.value = node.value
            return

        self.value = node

    def __repr__(self) -> str:
        """Return the string representation of the current node

        Returns:

            A string that describes the node

        Examples:

            Get the string representation of some example node ids

            >>> NodeId(0)
            Broadcast With Acknowledgment

            >>> NodeId(31)
            Broadcast Without Acknowledgment

            >>> NodeId(10)
            STH 10

            >>> NodeId(15)
            SPU 1

            >>> NodeId(18)
            STU 2

        """

        if self.value in (0, 31):
            return (
                "Broadcast "
                f"With{'' if self.value == 0 else 'out'} Acknowledgment"
            )
        if self.is_sth():
            return f"STH {self.value}"

        if 15 <= self.value <= 16:
            return f"SPU {self.value - 14}"

        return f"STU {self.value - 16}"

    def is_sth(self) -> bool:
        """Check if this node is an STH or not

        Returns:

            ``True`` if this node is an STH or ``False`` otherwise

        Examples:

            Check for some example node ids if they represent an STH or not

            >>> NodeId('STH 1').is_sth()
            True

            >>> NodeId('STU 12').is_sth()
            False

            >>> NodeId('STH 12').is_sth()
            True

        """

        return 1 <= self.value <= 14

    def is_stu(self) -> bool:
        """Check if this node is an STU or not

        Returns:

            ``True`` if this node is an STU or ``False`` otherwise

        Examples:

            Check for some example node ids if they represent an STU or not

            >>> NodeId('STH 7').is_stu()
            False

            >>> NodeId('STU 14').is_stu()
            True

            >>> NodeId('SPU 1').is_stu()
            False

        """

        return 17 <= self.value <= 30


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Utility functions for working with log data"""

# -- Imports ------------------------------------------------------------------

from logging import FileHandler, Formatter
from platformdirs import user_log_path

from icotronic.config import ConfigurationUtility

# -- Functions ----------------------------------------------------------------


def get_log_file_handler(filename: str) -> FileHandler:
    """Get file log handler that stores data in user’s log directory

    Args:

        filename:
            The filename of the log file that should be stored in the user’s
            log directory

    Examples:

        Initialize test data

        >>> from logging import getLogger
        >>> from platform import system
        >>> from re import match
        >>> from sys import platform

        >>> filename = "test.log"
        >>> log_filepath = user_log_path(
        ...     appname=ConfigurationUtility.app_name,
        ...     appauthor=ConfigurationUtility.app_author) / filename

        Workaround for missing support to remove files while “in use” on
        Windows (aka “The process cannot access the file because it is being
        used by another process”)

        >>> if system() == "Windows":
        ...     log_filepath.unlink(missing_ok=True)

        Initialize logger

        >>> logger = getLogger()
        >>> logger.addHandler(get_log_file_handler(filename))

        The log file should not exist until we add something to it

        >>> log_filepath = user_log_path(
        ...     appname=ConfigurationUtility.app_name,
        ...     appauthor=ConfigurationUtility.app_author) / "test.log"
        >>> log_filepath.exists()
        False

        The log file should exist after we add log information

        >>> logger.error("We are all gonna die")
        >>> log_filepath.exists()
        True

        The log file should contain the relevant logging information

        >>> log_content = open(log_filepath, encoding="utf8").readlines()
        >>> len(log_content)
        1
        >>> "We are all gonna die" in log_content[0]
        True

        Remove test log file

        >>> if system() != "Windows":
        ...     log_filepath.unlink()

    """

    log_filepath = (
        user_log_path(
            appname=ConfigurationUtility.app_name,
            appauthor=ConfigurationUtility.app_author,
        )
        / filename
    )

    # Create log directory, if it does not exist already
    if not log_filepath.parent.exists():
        log_filepath.parent.mkdir(
            exist_ok=True,
            parents=True,
        )

    handler = FileHandler(log_filepath, "w", "utf-8", delay=True)
    handler.setFormatter(Formatter("{asctime} {message}", style="{"))

    return handler


if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Convert MAC address to Base64 text"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser

from icotronic.cmdline.parse import mac_address
from icotronic.utility.naming import convert_mac_base64

# -- Main ---------------------------------------------------------------------


def main():
    """Convert MAC address specified via command line argument to Base64"""

    parser = ArgumentParser(
        description="Convert a MAC address to an 8 character Base64 text"
    )
    parser.add_argument(
        "mac",
        help="MAC address of STH e.g. 08:6b:d7:01:de:81",
        type=mac_address,
    )
    mac = parser.parse_args().mac

    name = convert_mac_base64(mac)
    print(name)


if __name__ == "__main__":
    main()



"""Data conversion functions"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterable

# -- Functions ----------------------------------------------------------------


def convert_bytes_to_text(
    data: Iterable[int], until_null: bool = False
) -> str:
    """Convert byte array data to a string

    Please note, that this function ignores non ASCII data and control
    characters

    Args:

        data:
            The byte array that should be converted

        until_null:
            Ignore data after the first ``NULL`` byte (``True``) or not
            ``False``

    Returns:

        An string where the (valid) bytes of the input are mapped to an ASCII
        character

    Examples:

        Convert some example byte array data into strings

        >>> convert_bytes_to_text("test".encode('ASCII'))
        'test'

        >>> input = bytearray([0, 255, 10, 30]) + "something".encode('ASCII')
        >>> convert_bytes_to_text(input)
        'something'

        >>> input = bytearray([0, 255, 10, 30]) + "something".encode('ASCII')
        >>> convert_bytes_to_text(input, until_null=True)
        ''

    """

    ascii_bytes = bytearray()
    for byte in data:
        if byte == 0 and until_null:
            break
        if ord(" ") < byte < 128:
            ascii_bytes.append(byte)

    return ascii_bytes.decode("ASCII")


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for the MyTooliT CAN protocol

See: https://mytoolit.github.io/Documentation/#mytoolit-communication-protocol

for more information
"""

# -- Exports ------------------------------------------------------------------

from .error import ErrorResponseError, CANConnectionError, NoResponseError
from .connection import Connection
from .streaming import (
    StreamingConfiguration,
    StreamingData,
    StreamingError,
    StreamingTimeoutError,
    StreamingBufferError,
)
from .node.sensor import SensorNode
from .node.stu import STU
from .node.sth import STH



"""Read some acceleration data of STH with node name Test-STH"""

# -- Imports ------------------------------------------------------------------

from asyncio import run

from icotronic.can import Connection, StreamingConfiguration

# -- Functions ----------------------------------------------------------------


async def read_streaming_data(identifier):
    """Read and output streaming data

    Args:

        identifier:

            Identifier of STH node

    """

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier) as sensor_node:
            # Read data of first channel
            async with sensor_node.open_data_stream(
                StreamingConfiguration(first=True)
            ) as stream:
                messages = 5
                async for data, _ in stream:
                    print(f"Read data values: {data}")
                    messages -= 1
                    if messages <= 0:
                        break


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    run(read_streaming_data(identifier="Test-STH"))



"""Support for sensor nodes (SHA, SMH and STH)"""

# pylint: disable=too-many-lines

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import CancelledError
from logging import getLogger
from types import TracebackType
from typing import NamedTuple

from netaddr import EUI

from icotronic.can.constants import (
    ADVERTISEMENT_TIME_EEPROM_TO_MS,
    SENSOR_NODE_NUMBER_SELF_ADDRESSING,
)
from icotronic.can.adc import ADCConfiguration
from icotronic.can.node.eeprom.sensor import SensorNodeEEPROM
from icotronic.can.error import (
    ErrorResponseError,
    NoResponseError,
    UnsupportedFeatureException,
)
from icotronic.can.protocol.message import Message
from icotronic.can.node.basic import Node
from icotronic.can.node.id import NodeId
from icotronic.can.streaming import (
    AsyncStreamBuffer,
    StreamingConfiguration,
    StreamingData,
    StreamingFormat,
    StreamingFormatVoltage,
)
from icotronic.can.node.spu import SPU
from icotronic.config import settings
from icotronic.measurement.sensor import SensorConfiguration
from icotronic.measurement.voltage import convert_raw_to_supply_voltage

# -- Classes ------------------------------------------------------------------


class DataStreamContextManager:
    """Open and close a data stream from a sensor node

    Args:

        sensor_node:
            The sensor node for which this context manager handles
            the streaming data

        channels:
            A streaming configuration that specifies which of the three
            streaming channels should be enabled or not

        timeout
            The amount of seconds between two consecutive messages, before
            a TimeoutError will be raised

    """

    def __init__(
        self,
        sensor_node: SensorNode,
        channels: StreamingConfiguration,
        timeout: float,
    ) -> None:

        self.node = sensor_node
        self.channels = channels
        self.timeout = timeout
        self.reader: AsyncStreamBuffer | None = None
        self.logger = getLogger()
        self.logger.debug("Initialized data stream context manager")

    async def __aenter__(self) -> AsyncStreamBuffer:
        """Open the stream of measurement data

        Returns:

            The stream buffer for the measurement stream

        """

        adc_config = await self.node.get_adc_configuration()
        # Raise exception if there if there is more than one second worth
        # of buffered data
        self.reader = AsyncStreamBuffer(
            self.channels,
            self.timeout,
            max_buffer_size=round(adc_config.sample_rate()),
        )

        self.node.spu.notifier.add_listener(self.reader)
        await self.node.start_streaming_data(self.channels)
        self.logger.debug("Entered data stream context manager")

        return self.reader

    async def __aexit__(
        self,
        exception_type: type[BaseException] | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Clean up the resources used by the stream

        Args:

            exception_type:
                The type of the exception in case of an exception

            exception_value:
                The value of the exception in case of an exception

            traceback:
                The traceback in case of an exception

        """

        if self.reader is not None:
            self.reader.stop()
            self.node.spu.notifier.remove_listener(self.reader)

        if exception_type is None or isinstance(
            exception_type, type(CancelledError)
        ):
            self.logger.info("Stopping stream")
            await self.node.stop_streaming_data()
        else:
            # If there was an error while streaming data, then stoping the
            # stream will usually also fail. Because of this we only try once
            # and ignore any errors.
            #
            # If we did not do that, then the user of the API would be notified
            # about the error to disable the stream, but not about the original
            # error. It would also take considerably more time until the
            # computer would report an error, since the code would usually try
            # to stop the stream (and fail) multiple times beforehand.
            self.logger.info(
                "Stopping stream after error (%s)", exception_type
            )
            await self.node.stop_streaming_data(retries=1, ignore_errors=True)


class Times(NamedTuple):
    """Advertisement time and time until deeper sleep mode"""

    advertisement: float
    """Advertisement time"""

    sleep: int
    """Time until deeper sleep mode"""

    def __repr__(self) -> str:
        """Return a string representation of the object

        Returns:

            A string that contains the advertisement time and sleep time values

        """

        return ", ".join([
            f"Advertisement Time: {self.advertisement} ms",
            f"Sleep Time: {self.sleep} ms",
        ])


class SensorNode(Node):
    """Communicate and control a connected sensor node (SHA, STH, SMH)

    Args:

        spu:
            The SPU object used to connect to this sensor node

        eeprom:
            The EEPROM class of the node

    """

    def __init__(
        self, spu: SPU, eeprom: type[SensorNodeEEPROM] = SensorNodeEEPROM
    ) -> None:

        super().__init__(spu, eeprom, NodeId("STH 1"))

    # ==========
    # = System =
    # ==========

    # -------------
    # - Bluetooth -
    # -------------

    async def get_name(self) -> str:
        """Retrieve the name of the sensor node

        Returns:

            The (Bluetooth broadcast) name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Get Bluetooth advertisement name of node “0”

            >>> async def get_sensor_node_name():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_name()
            >>> name = run(get_sensor_node_name())
            >>> isinstance(name, str)
            True
            >>> 0 <= len(name) <= 8
            True

        """

        return await self.spu.get_name(
            node=self.id, sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING
        )

    async def set_name(self, name: str) -> None:
        """Set the name of a sensor node

        Args:

            name:
                The new name for the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Change the name of a sensor node

            >>> async def test_naming(name):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         # and that this node currently does not have the name
            ...         # specified in the variable `name`.
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             before = await sensor_node.get_name()
            ...             await sensor_node.set_name(name)
            ...             updated = await sensor_node.get_name()
            ...             await sensor_node.set_name(before)
            ...             after = await sensor_node.get_name()
            ...             return before, updated, after
            >>> before, updated, after = run(test_naming("Hello"))
            >>> before != "Hello"
            True
            >>> updated
            'Hello'
            >>> before == after
            True

        """

        if not isinstance(name, str):
            raise TypeError("Name must be str, not type(identifier).__name__")

        bytes_name = list(name.encode("utf-8"))
        length_name = len(bytes_name)
        if length_name > 8:
            raise ValueError(
                f"Name is too long ({length_name} bytes). "
                "Please use a name between 0 and 8 bytes."
            )

        node = self.id
        # Use 0 bytes at end of names that are shorter than 8 bytes
        bytes_name.extend([0] * (8 - length_name))
        description = f"name of “{node}”"

        await self.spu.request_bluetooth(
            node=node,
            subcommand=3,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            data=bytes_name[:6],
            description=f"set first part of {description}",
        )

        await self.spu.request_bluetooth(
            node=node,
            subcommand=4,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            data=bytes_name[6:] + [0] * 4,
            description=f"set second part of {description}",
        )

    async def get_rssi(self):
        """Retrieve the RSSI (Received Signal Strength Indication) of the node

        Returns:

            The RSSI of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Get RSSI of node “0”

            >>> async def get_sensor_node_rssi():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_rssi()
            >>> rssi = run(get_sensor_node_rssi())
            >>> -70 < rssi < 0
            True

        """

        return await self.spu.get_rssi(
            self.id, SENSOR_NODE_NUMBER_SELF_ADDRESSING
        )

    async def get_energy_mode_reduced(self) -> Times:
        """Read the reduced energy mode (mode 1) sensor node time values

        See also:

        - https://mytoolit.github.io/Documentation/#sleep-advertisement-times

        Returns:

            A tuple containing the advertisement time in the reduced energy
            mode in milliseconds and the time until the node will switch from
            the disconnected state to the low energy mode (mode 1) – if there
            is no activity – in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Retrieve the reduced energy time values of a sensor node

            >>> async def read_energy_mode_reduced():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_energy_mode_reduced()
            >>> times = run(read_energy_mode_reduced())
            >>> round(times.advertisement)
            1250
            >>> times.sleep
            300000

        """

        response = await self.spu.request_bluetooth(
            node=self.id,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            subcommand=13,
            description="get reduced energy time values of sensor node",
        )

        wait_time = int.from_bytes(response.data[2:6], byteorder="little")
        advertisement_time = (
            int.from_bytes(response.data[6:], byteorder="little")
            * ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        return Times(sleep=wait_time, advertisement=advertisement_time)

    async def set_energy_mode_reduced(
        self, times: Times | None = None
    ) -> None:
        """Writes the time values for the reduced energy mode (mode 1)

        See also:

        - https://mytoolit.github.io/Documentation/#sleep-advertisement-times

        Args:

            times:
                The values for the advertisement time in the reduced energy
                mode in milliseconds and the time until the node will go into
                the low energy mode (mode 1) from the disconnected state – if
                there is no activity – in milliseconds.

                If you do not specify these values then the default values
                from the configuration will be used

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read and write the reduced energy time values of a sensor node

            >>> async def read_write_energy_mode_reduced(sleep, advertisement):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.set_energy_mode_reduced(
            ...                 Times(sleep=sleep,
            ...                       advertisement=advertisement))
            ...             times = await sensor_node.get_energy_mode_reduced()
            ...
            ...             # Overwrite changed values with default config
            ...             # values
            ...             await sensor_node.set_energy_mode_reduced()
            ...
            ...             return times
            >>> times = run(read_write_energy_mode_reduced(200_000, 2000))
            >>> times.sleep
            200000
            >>> round(times.advertisement)
            2000

        """

        if times is None:
            time_settings = settings.sensor_node.bluetooth
            times = Times(
                sleep=time_settings.sleep_time_1,
                advertisement=time_settings.advertisement_time_1,
            )

        sleep_time = times.sleep
        advertisement_time = round(
            times.advertisement / ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        data = list(
            sleep_time.to_bytes(4, "little")
            + advertisement_time.to_bytes(2, "little")
        )

        await self.spu.request_bluetooth(
            node=self.id,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            subcommand=14,
            data=data,
            response_data=list(data),
            description="set reduced energy time values of sensor node",
        )

    async def get_energy_mode_lowest(self) -> Times:
        """Read the reduced lowest energy mode (mode 2) time values

        See also:

        - https://mytoolit.github.io/Documentation/#sleep-advertisement-times

        Returns:

            A tuple containing the advertisement time in the lowest energy
            mode in milliseconds and the time until the node will switch from
            the reduced energy mode (mode 1) to the lowest energy mode (mode
            2) – if there is no activity – in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Retrieve the reduced energy time values of a sensor node

            >>> async def read_energy_mode_lowest():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_energy_mode_lowest()
            >>> times = run(read_energy_mode_lowest())
            >>> round(times.advertisement)
            2500
            >>> times.sleep
            259200000

        """

        response = await self.spu.request_bluetooth(
            node=self.id,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            subcommand=15,
            description="get lowest energy mode time values of sensor node",
        )

        wait_time = int.from_bytes(response.data[2:6], byteorder="little")
        advertisement_time = (
            int.from_bytes(response.data[6:], byteorder="little")
            * ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        return Times(sleep=wait_time, advertisement=advertisement_time)

    async def set_energy_mode_lowest(self, times: Times | None = None) -> None:
        """Writes the time values for the lowest energy mode (mode 2)

        See also:

        - https://mytoolit.github.io/Documentation/#sleep-advertisement-times

        Args:

            times:
                The values for the advertisement time in the reduced energy
                mode in milliseconds and the time until the node will go into
                the lowest energy mode (mode 2) from the reduced energy mode
                (mode 1) – if there is no activity – in milliseconds.

                If you do not specify these values then the default values
                from the configuration will be used

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read and write the reduced energy time values of a sensor node

            >>> async def read_write_energy_mode_lowest(sleep, advertisement):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.set_energy_mode_lowest(
            ...                 Times(sleep=sleep,
            ...                       advertisement=advertisement))
            ...             times = await sensor_node.get_energy_mode_lowest()
            ...
            ...             # Overwrite changed values with default config
            ...             # values
            ...             await sensor_node.set_energy_mode_lowest()
            ...
            ...             return times
            >>> times = run(read_write_energy_mode_lowest(200_000, 2000))
            >>> times.sleep
            200000
            >>> round(times.advertisement)
            2000

        """

        if times is None:
            time_settings = settings.sensor_node.bluetooth
            times = Times(
                sleep=time_settings.sleep_time_2,
                advertisement=time_settings.advertisement_time_2,
            )

        sleep_time = times.sleep
        advertisement_time = round(
            times.advertisement / ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        data = list(
            sleep_time.to_bytes(4, "little")
            + advertisement_time.to_bytes(2, "little")
        )

        await self.spu.request_bluetooth(
            node=self.id,
            sensor_node_number=SENSOR_NODE_NUMBER_SELF_ADDRESSING,
            subcommand=16,
            data=data,
            response_data=list(data),
            description="set reduced energy time values of sensor node",
        )

    async def get_mac_address(self) -> EUI:
        """Retrieve the MAC address of the sensor node

        Returns:

            The MAC address of the specified sensor node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Retrieve the MAC address of STH 1

            >>> async def get_bluetooth_mac():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_mac_address()
            >>> mac_address = run(get_bluetooth_mac())
            >>> isinstance(mac_address, EUI)
            True
            >>> mac_address != EUI(0)
            True

        """

        return await self.spu.get_mac_address(
            self.id, SENSOR_NODE_NUMBER_SELF_ADDRESSING
        )

    # =============
    # = Streaming =
    # =============

    # --------
    # - Data -
    # --------

    async def get_streaming_data_single(
        self,
        channels=StreamingConfiguration(first=True, second=True, third=True),
    ) -> StreamingData:
        """Read a single set of raw ADC values from the sensor node

        Args:

            channels:
                Specifies which of the three measurement channels should be
                enabled or disabled

        Returns:

            The latest three ADC values measured by the sensor node

        Examples:

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read a single value from all three measurement channels

            >>> async def read_sensor_values():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return (await
            ...                 sensor_node.get_streaming_data_single())
            >>> data = run(read_sensor_values())
            >>> len(data.values)
            3
            >>> all([0 <= value <= 0xffff for value in data.values])
            True

        """

        streaming_format = StreamingFormat(
            channels=channels,
            sets=1,
        )

        node = self.id

        response = await self.spu.request(
            Message(
                block="Streaming",
                block_command="Data",
                sender=self.spu.id,
                receiver=self.id,
                request=True,
                data=[streaming_format.value],
            ),
            description=f"read single set of streaming values from “{node}”",
        )

        values = [
            int.from_bytes(word, byteorder="little")
            for word in (
                response.data[2:4],
                response.data[4:6],
                response.data[6:8],
            )
        ]
        assert len(values) == 2 or len(values) == 3

        data = StreamingData(
            values=values,
            timestamp=response.timestamp,
            counter=response.data[1],
        )

        return data

    async def start_streaming_data(
        self, channels: StreamingConfiguration
    ) -> None:
        """Start streaming data

        Args:

            channels:
                Specifies which of the three measurement channels should be
                enabled or disabled

        The CAN identifier that this coroutine returns can be used
        to filter CAN messages that contain the expected streaming data

        """

        streaming_format = StreamingFormat(
            channels=channels,
            streaming=True,
            sets=3 if channels.enabled_channels() <= 1 else 1,
        )
        node = self.id
        message = Message(
            block="Streaming",
            block_command="Data",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[streaming_format.value],
        )

        measurement_channels = [
            channel
            for channel in (
                "first" if channels.first else "",
                "second" if channels.second else "",
                "third" if channels.third else "",
            )
            if channel
        ]
        channels_text = "".join(
            (f"{channel}, " for channel in measurement_channels[:-2])
        ) + " and ".join(measurement_channels[-2:])

        await self.spu.request(
            message,
            description=(
                f"enable streaming of {channels_text} measurement "
                f"channel of “{node}”"
            ),
        )

    async def stop_streaming_data(
        self, retries: int = 10, ignore_errors=False
    ) -> None:
        """Stop streaming data

        Args:

            retries:
                The number of times the message is sent again, if no response
                was sent back in a certain amount of time

            ignore_errors:
                Specifies, if this coroutine should ignore, if there were any
                problems while stopping the stream.

        """

        streaming_format = StreamingFormat(streaming=True, sets=0)
        node = self.id
        message = Message(
            block="Streaming",
            block_command="Data",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[streaming_format.value],
        )

        try:

            await self.spu.request(
                message,
                description=f"disable data streaming of “{node}”",
                retries=retries,
            )

        except (NoResponseError, ErrorResponseError) as error:
            if not ignore_errors:
                raise error

    def open_data_stream(
        self,
        channels: StreamingConfiguration,
        timeout: float = 5,
    ) -> DataStreamContextManager:
        """Open measurement data stream

        Args:

            channels:
                Specifies which measurement channels should be enabled

            timeout:
                The amount of seconds between two consecutive messages, before
                a TimeoutError will be raised

        Returns:

            A context manager object for managing stream data

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read data of first and third channel

            >>> async def read_streaming_data():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             channels = StreamingConfiguration(first=True,
            ...                                               third=True)
            ...             async with sensor_node.open_data_stream(
            ...               channels) as stream:
            ...                 first = []
            ...                 third = []
            ...                 messages = 0
            ...                 async for data, _ in stream:
            ...                     first.append(data.values[0])
            ...                     third.append(data.values[1])
            ...                     messages += 1
            ...                     if messages >= 3:
            ...                         break
            ...                 return first, third
            >>> first, third = run(read_streaming_data())
            >>> len(first)
            3
            >>> len(third)
            3

        """

        return DataStreamContextManager(self, channels, timeout)

    # -----------
    # - Voltage -
    # -----------

    async def get_supply_voltage(self) -> float:
        """Read the current supply voltage

        Returns:

            The supply voltage of the sensor node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the supply voltage of the sensor node with node number 0

            >>> async def get_supply_voltage():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_supply_voltage()
            >>> supply_voltage = run(get_supply_voltage())
            >>> 3 <= supply_voltage <= 4.2
            True

        """

        streaming_format = StreamingFormatVoltage(
            channels=StreamingConfiguration(first=True), sets=1
        )
        node = self.id
        message = Message(
            block="Streaming",
            block_command="Voltage",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[streaming_format.value],
        )

        response = await self.spu.request(
            message, description=f"read supply voltage of “{node}”"
        )

        voltage_bytes = response.data[2:4]
        voltage_raw = int.from_bytes(voltage_bytes, "little")

        adc_configuration = await self.get_adc_configuration()

        return convert_raw_to_supply_voltage(
            voltage_raw,
            reference_voltage=adc_configuration.reference_voltage,
        )

    # =================
    # = Configuration =
    # =================

    # -----------------------------
    # - Get/Set ADC Configuration -
    # -----------------------------

    async def get_adc_configuration(self) -> ADCConfiguration:
        """Read the current ADC configuration

        Returns:

            The ADC configuration of the sensor node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read ADC sensor config of sensor node with node id 0

            >>> async def read_adc_config():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_adc_configuration()
            >>> run(read_adc_config()) # doctest:+NORMALIZE_WHITESPACE
            Get, Prescaler: 2, Acquisition Time: 8, Oversampling Rate: 64,
            Reference Voltage: 3.3 V

        """

        node = self.id

        message = Message(
            block="Configuration",
            block_command="Get/Set ADC Configuration",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[0] * 8,
        )

        response = await self.spu.request(
            message, description=f"Read ADC configuration of “{node}”"
        )

        return ADCConfiguration(response.data[0:5])

    async def set_adc_configuration(
        self,
        reference_voltage: float = 3.3,
        prescaler: int = 2,
        acquisition_time: int = 8,
        oversampling_rate: int = 64,
    ) -> None:
        """Change the ADC configuration of a connected sensor node

        Args:

            reference_voltage:
                The ADC reference voltage in Volt
                (1.25, 1.65, 1.8, 2.1, 2.2, 2.5, 2.7, 3.3, 5, 6.6)

            prescaler:
                The ADC prescaler value (1 – 127)

            acquisition_time:
                The ADC acquisition time in number of cycles
                (1, 2, 3, 4, 8, 16, 32, … , 256)

            oversampling_rate:
                The ADC oversampling rate (1, 2, 4, 8, … , 4096)

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read and write ADC sensor config

            >>> async def write_read_adc_config():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.set_adc_configuration(
            ...                 3.3, 8, 8, 64)
            ...             modified_config1 = (await
            ...                 sensor_node.get_adc_configuration())
            ...
            ...             adc_config = ADCConfiguration(
            ...                 reference_voltage=5.0,
            ...                 prescaler=16,
            ...                 acquisition_time=8,
            ...                 oversampling_rate=128)
            ...             await sensor_node.set_adc_configuration(
            ...                 **adc_config)
            ...             modified_config2 = (await
            ...                 sensor_node.get_adc_configuration())
            ...
            ...             # Write back default config values
            ...             await sensor_node.set_adc_configuration(
            ...                 3.3, 2, 8, 64)
            ...             return modified_config1, modified_config2
            >>> config1, config2 = run(write_read_adc_config())
            >>> config1 # doctest:+NORMALIZE_WHITESPACE
            Get, Prescaler: 8, Acquisition Time: 8, Oversampling Rate: 64,
            Reference Voltage: 3.3 V
            >>> config2 # doctest:+NORMALIZE_WHITESPACE
            Get, Prescaler: 16, Acquisition Time: 8, Oversampling Rate: 128,
            Reference Voltage: 5.0 V

        """

        node = self.id
        adc_configuration = ADCConfiguration(
            set=True,
            prescaler=prescaler,
            acquisition_time=acquisition_time,
            oversampling_rate=oversampling_rate,
            reference_voltage=reference_voltage,
        )

        message = Message(
            block="Configuration",
            block_command="Get/Set ADC Configuration",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=adc_configuration.data,
        )

        await self.spu.request(
            message, description=f"write ADC configuration of “{node}”"
        )

    # --------------------------------
    # - Get/Set Sensor Configuration -
    # --------------------------------

    async def get_sensor_configuration(self) -> SensorConfiguration:
        """Read the current sensor configuration

        Raises:

            UnsupportedFeatureException
                in case the sensor node replies with an error message

        Returns:

            The sensor number for the different axes

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Reading sensor config from node without sensor config support fails

            >>> async def read_sensor_config():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.get_sensor_configuration()
            >>> config = run(
            ...     read_sensor_config()) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
               ...
            UnsupportedFeatureException: Reading sensor configuration is not
            supported

        """

        node = self.id
        message = Message(
            block="Configuration",
            block_command=0x01,
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[0] * 8,
        )

        try:

            response = await self.spu.request(
                message, description=f"get sensor configuration of “{node}”"
            )

        except ErrorResponseError as error:
            raise UnsupportedFeatureException(
                "Reading sensor configuration not supported"
            ) from error

        channels = response.data[1:4]

        return SensorConfiguration(*channels)

    async def set_sensor_configuration(
        self, sensors: SensorConfiguration
    ) -> None:
        """Change the sensor numbers for the different measurement channels

        If you use the sensor number `0` for one of the different measurement
        channels, then the sensor (number) for that channel will stay the same.

        Args:

            sensors:
                The sensor numbers of the different measurement channels

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Setting sensor config from node without sensor config support fails

            >>> async def set_sensor_config():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.set_sensor_configuration(
            ...                 SensorConfiguration(
            ...                     first=0, second=0, third=0))
            >>> config = run(
            ...     set_sensor_config()) #doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
               ...
            UnsupportedFeatureException: Writing sensor configuration is not
            supported

        """

        node = self.id
        data = [
            0b1000_0000,
            sensors.first,
            sensors.second,
            sensors.third,
            *(4 * [0]),
        ]
        message = Message(
            block="Configuration",
            block_command=0x01,
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=data,
        )

        try:

            await self.spu.request(
                message, description=f"set sensor configuration of “{node}”"
            )

        except ErrorResponseError as error:
            raise UnsupportedFeatureException(
                "Writing sensor configuration not supported"
            ) from error


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Read name of sensor node with node number 0"""

# -- Imports ------------------------------------------------------------------

from asyncio import run

from netaddr import EUI

from icotronic.can import Connection

# -- Functions ----------------------------------------------------------------


async def read_name(identifier: EUI | str | int) -> None:
    """Read sensor node name

    Args:

        identifier:

            Identifier of STH node

    """

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier) as sensor_node:
            name = await sensor_node.get_name()
            print(f"Connected to sensor node “{name}”")


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    # Possible Identifiers:
    # - Name:               e.g. `"Test-STH"`
    # - Sensor Node Number: e.g. `1`
    # - MAC Address:        e.g. `netaddr.EUI('08-6B-D7-01-DE-81')`
    run(read_name(identifier=0))
